/// Type alias for json object map
typedef JsonObject = Map<String, dynamic>;

/// Converts a [JsonObject] into [T]
typedef FromJsonObject<T> = T Function(JsonObject json);

/// Converts a json element into [T]
typedef FromJson<T> = T Function(dynamic json);

/// Mixin for Data Transfer Objects
/// Every subclass of this mixin must also provide fromJson static method:
/// ```dart
/// static Foo fromJson(JsonObject json) => Foo(name: json['name'], age: json['age']);
/// ```
mixin DTO {
  /// Converts this instance into a json object map. E.g. JsonObject toJson() => {'name': name, 'age': age};
  JsonObject toJson();

  /// Default toString implementation: E.g. Foo(name: Ugo, age: 37)
  @override
  String toString() => '$runtimeType(${toJson().entries.map((entry) => '${entry.key}: ${entry.value}').join(', ')})';
}

/// Shortcut for json['xx'] == null ? null : fromJson(json['xx'])
/// Usage: fromNullableJson(json['xx'], fromJson)
T? fromNullableJson<T>(JsonObject? json, FromJsonObject<T> fromJson) => json == null ? null : fromJson(json);

/// Shortcut for (json['xx'] as List).map((json) => fromJson(json)).toList(growable: false)
/// Usage: fromJsonList(json['xx'], fromJson)
List<T> fromJsonList<T>(Iterable json, FromJsonObject<T> fromJson, {bool growable = false}) =>
    List<T>.of(json.map((json) => fromJson(json)), growable: growable);

/// Shortcut for json['xx'] == null ? null : fromJsonList(json['xx'], fromJson)
/// Usage: fromNullableJsonList(json['xx'], fromJson)
List<T>? fromNullableJsonList<T>(Iterable? json, FromJsonObject<T> fromJson, {bool growable = false}) =>
    json == null ? null : fromJsonList(json, fromJson, growable: growable);

/// Shortcut for (json['xx'] as Map).map((key, value) => MapEntry(key, fromJson(value)))
/// Usage: fromJsonMap(json['xx'], fromJson)
Map<String, T> fromJsonMap<T>(Map json, FromJsonObject<T> fromJson) =>
    json.map<String, T>((key, value) => MapEntry(key, fromJson(value)));

/// Shortcut for json['xx'] == null ? null : fromJsonMap(json['xx'], fromJson)
/// Usage: fromNullableJsonMap(json['xx'], fromJson)
Map<String, T>? fromNullableJsonMap<T>(Map? json, FromJsonObject<T> fromJson) =>
    json == null ? null : fromJsonMap(json, fromJson);

extension DTOListExt<T extends DTO> on Iterable<T> {
  /// Shortcut for items.map((item) => item.toJson()).toList(growable: false)
  /// Usage: items.toJsonList()
  List<JsonObject> toJsonList({bool growable = false}) =>
      List<JsonObject>.of(map((item) => item.toJson()), growable: growable);
}
