import 'dart:convert';
import 'dart:developer' as dev;

import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';

/// Logging utility
mixin Loggable {
  static final Map<String, LogLevel> _levelByTag = {};
  String? _tag;

  /// [Indent] of the log
  final Indent indent = Indent(4);

  /// Log tag which is the runtime type of who implements this mixin by default
  String get tag => _tag ??= tagOf(this);

  /// Log level
  LogLevel get level => _levelByTag[tag] ?? LogLevel.debug;

  void logDebug(Object message, [VoidCallback? details]) => logInternal(LogLevel.debug, message, details: details);

  void logInfo(Object message, [VoidCallback? details]) => logInternal(LogLevel.info, message, details: details);

  void logWarning(Object message, [VoidCallback? details]) => logInternal(LogLevel.warning, message, details: details);

  void logError(Object message, [Object? error, StackTrace? stack, VoidCallback? details]) =>
      logInternal(LogLevel.error, message, error: error, stack: stack, details: details);

  void log(LogLevel level, Object message, {Object? error, StackTrace? stack, VoidCallback? details}) =>
      logInternal(level, message, error: error, stack: stack, details: details);

  /// Actual log implementation, override this method to handle log write
  @protected
  void logInternal(LogLevel level, Object message, {Object? error, StackTrace? stack, VoidCallback? details}) {
    assert(() {
      if (this.level.index > level.index) return true;
      if (message is Printable) {
        message.print(this);
        return true;
      }
      final lines = LineSplitter.split(message.toString());
      final name = tag;
      var prefix = '${describeEnum(level).toUpperCase()} - $indent';
      var count = 0;
      final last = lines.length - 1;
      for (final line in lines) {
        if (count == 1) prefix = ' ' * prefix.length;
        if (count == last)
          _logLineInternal('$prefix$line', name, error, stack);
        else
          _logLineInternal('$prefix$line', name, null, null);
        count++;
      }
      if (details != null)
        try {
          indent.increase();
          details();
        } finally {
          indent.decrease();
        }
      return true;
    }());
  }

  void _logLineInternal(String line, String name, Object? error, StackTrace? stack) =>
      dev.log(line, name: name, error: error, stackTrace: stack);

  static void setLevelFor<T>(LogLevel level) => setLevelForTag(T.toString(), level);

  static void setLevelForTag(String tag, LogLevel level) => _levelByTag[tag] = level;

  static String tagOf(Object object) => (object is State ? object.widget.runtimeType : object.runtimeType).toString();
}

enum LogLevel { debug, info, warning, error, none }

/// Utility class that helps to provide indent spacing
/// Just through [toString] you can get the indentation of current level
class Indent {
  int _level = 0;

  /// The number of spaces per indent level
  final int step;

  /// The current indent level can be modified by invoking [increase] and [decrease]
  int get level => _level;

  /// Create Indent instance with given [step]
  Indent(this.step);

  /// Get indent by [level]
  String operator [](int level) => ' ' * (step * level);

  /// Increase level
  void increase() => _level++;

  /// Decrease level
  void decrease() => _level--;

  @override
  String toString() => this[_level];
}

/// Implement this interface to provide specific logging for an object
/// [print] is called when you pass this instance as message argument of [Loggable.log] e.g. log(myPrintable)
abstract class Printable {
  const Printable();

  /// Describe this instance using [_] [Loggable]
  /// Waring: calling _.log(this) will result in a stack overflow
  @protected
  void print(Loggable _);
}
