import 'dart:async';

import 'package:flutter/foundation.dart';

/// Mutual exclusion.
/// Acquire and release lock
class Mutex {
  Completer<void>? _completer;

  bool get running => _completer != null;

  Future<T> lock<T>(Future<T> Function() function) => acquire().then((value) async {
        try {
          return await function();
        } finally {
          release();
        }
      });

  Future<void> acquire() {
    final c = _completer;
    if (c == null) {
      _completer = Completer();
      return SynchronousFuture(null);
    } else {
      return c.future.then((_) => acquire());
    }
  }

  void release() {
    final c = _completer;
    if (c == null) throw StateError('cannot release, nothing acquired');
    c.complete();
    _completer = null;
  }
}

/// Run async function successfully exactly once
class RunOnce extends Mutex {
  bool _done = false;

  Future<bool> call(Future Function() function) => _done || running
      ? SynchronousFuture(false)
      : acquire().then((value) async {
          try {
            if (_done) return false;
            await function();
            return _done = true;
          } finally {
            release();
          }
        });
}

/// Run async function after [delay] only if [call] has never been called twice in that delay
class RunDelayed extends Mutex {
  final Duration delay;
  late DateTime _lastCall;

  RunDelayed(this.delay);

  Future<bool> call(Future Function() function) {
    final now = _lastCall = DateTime.now();
    return Future.delayed(
        delay,
        () => now == _lastCall
            ? acquire().then((value) async {
                try {
                  if (now != _lastCall) return false;
                  await function();
                  return true;
                } finally {
                  release();
                }
              })
            : SynchronousFuture(false));
  }
}
